import os
import click
import tempfile
import subprocess
import shutil

def write(name, targetdir, action, verbose):
    # XXX: insecure
    if verbose:
        click.echo('write to %(target)s' % action)
    target = os.path.join(targetdir, action['target'])
    targetdir = os.path.dirname(target)
    if not os.path.isdir(targetdir):
        os.makedirs(targetdir)
    with open(target, 'w') as fp:
        fp.write(action['content'])


def shell(name, targetdir, action, verbose):


    if action['cwd']:
        # the cwd option is insecure and used for examples
        # that already have all files in place
        # like an examples folder for example

        src = os.path.join(
            os.path.abspath(
                os.path.dirname(action['file'])),
            action['cwd'])
        targetdir = os.path.join(targetdir, action['cwd'])
        shutil.copytree(src, targetdir)

    if not os.path.isdir(targetdir):
        os.makedirs(targetdir)

    if verbose:
        click.echo('popen %(target)s' % action)
    proc = subprocess.Popen(
        action['target'],
        shell=True,
        cwd=targetdir,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=0,
    )
    out, err = proc.communicate()
    out = out.decode('utf-8')
    assert not err
    return out



def wipe(name, targetdir, action, verbose):
    if verbose:
        click.secho('wiping targetdir %s of %s' % (targetdir, name), bold=True)
    shutil.rmtree(targetdir)
    os.mkdir(targetdir)


ACTIONS = {
    'shell': shell,
    'wipe': wipe,
    'write': write,
}



import os
import sys
import click
import tempfile
import subprocess
import shutil
import re

from .parse import parse_actions, correct_content
from .actions import ACTIONS


def normalize_content(content, operators):
    lines = content.splitlines(True)
    result = []
    for line in lines:
        for op in operators:
            line = op(line)
        result.append(line)
    return ''.join(result)


def check_file(name, content, tmpdir, normalize, verbose=True):
    needed_updates = []
    for action in parse_actions(content, file=name):
        method = ACTIONS[action['action']]
        new_content = method(
            name=name,
            targetdir=tmpdir,
            action=action,
            verbose=verbose)
        if new_content:
            action['new_content'] = normalize_content(new_content, normalize)
            needed_updates.append(action)
    return needed_updates


def print_diff(action):
    content, out = action['content'], action['new_content']
    if out != content:
        import difflib
        differ = difflib.Differ()
        outl = out.splitlines(True)
        contl = content.splitlines(True)
        lines = differ.compare(contl, outl)

        mapping = {
            '+': 'green',
            '-': 'red',
            '?': 'blue',
        }
        if lines:
            click.secho("$ " + action['target'], bold=True, fg='blue')
        for line in lines:
            color = mapping.get(line[0])
            if color:
                click.secho(line, fg=color, bold=True, nl=False)
            else:
                click.echo(line, nl=False)


class Substituter(object):
    def __init__(self, match, replace):
        self.match = match
        self.replace = replace

    @classmethod
    def parse(cls, s):
        parts = s.split(s[0])
        assert len(parts) == 4
        return cls(
            match=parts[1],
            replace=parts[2],
        )

    def __call__(self, line):
        return re.sub(self.match, self.replace, line)

    def __repr__(self):
        return '<Substituter {self.match!r} to {self.replace!r}>'.format(
            self=self)




@click.command()
@click.argument('files', nargs=-1)
@click.option('--update', is_flag=True)
@click.option('--normalize', type=Substituter.parse, multiple=True)
@click.option('--verbose', default=False, is_flag=True)
def main(files, update, normalize=(), rootdir=None, verbose=False):
    tmpdir = rootdir or tempfile.mkdtemp(prefix='regendoc-exec-')
    total = len(files)
    for num, name in enumerate(files, 1):
        targetdir = os.path.join(tmpdir, '%s-%d' % (
            os.path.basename(name), num))
        with open(name) as fp:
            content = list(fp)
        os.mkdir(targetdir)
        click.secho(
            '#[{num:3d}/{total:3d}] {name}'.format(
                num=num, total=total, name=name), bold=True)
        updates = check_file(
            name=name,
            content=content,
            tmpdir=targetdir,
            normalize=(
                Substituter(
                    match=re.escape(targetdir),
                    replace='$REGENDOC_TMPDIR',
                ),
                Substituter(
                    match=re.escape(os.getcwd()),
                    replace='$PWD',
                ),
                Substituter(
                    match='at 0x[0-9a-f]+>',
                    replace='at 0xdeadbeef>',
                )
            ) + normalize,
            verbose=verbose,
        )
        for action in updates:
            if action['content'] is None or action['new_content'] is None:
                continue
            print_diff(action)
        if update:
            corrected = correct_content(content, updates)
            with open(name, "w") as f:
                f.writelines(corrected)



import textwrap
import pytest

from regendoc.parse import(
    blocks, correct_content,
    classify, parse_actions,
)

from regendoc import (
    main,
    check_file,
)

from operator import itemgetter
a_c_t = itemgetter('action', 'target')


@pytest.fixture
def run():
    def _run(*args, **kw):
        newkw = dict((k, str(v) if not isinstance(v, bool) else v)
                     for k, v in kw.items())
        args = [str(x) for x in args]
        main.callback(args, **newkw)
    return _run


input_data_for_blocks = """
some text
more text

next block::

    indent
      block

    with more

.. directive:: test
    :param: test

    text
"""


simple = """
an echo:
    $ echo hi
    oh no
"""

simple_corrected = simple.replace('oh no', 'hi')


def test_blocks():
    result = blocks(input_data_for_blocks.splitlines(True))
    expected = [
        # indent level, start line, lines
        (0, 1, ['some text\n', 'more text\n', '\n', 'next block::\n']),
        (4, 6, ['indent\n', '  block\n', '\n', 'with more\n']),
        (0, 11, ['.. directive:: test\n']),
        (4, 12, [':param: test\n', '\n', 'text\n']),
    ]
    assert result == expected


def test_classify_write():

    write = classify([
        '# content of test_foo.py\n',
        'def test()\n',
        '    pass\n',
    ])

    expected = {
        'action': 'write',
        'cwd': None,
        'target': 'test_foo.py',
        'content': 'def test()\n    pass\n',
        'indent': 4,
        'line': None,
    }
    assert write == expected


def test_classify_shell():
    cmd = classify(lines=[
        '$ py.test -x\n',
        'crud\n',
    ])
    expected = {
        'action': 'shell',
        'cwd': None,
        'target': 'py.test -x',
        'content': 'crud\n',
        'indent': 4,
        'line': None,
    }
    assert cmd == expected


def test_classify_chdir_shell():
    cmd = classify(lines=[
        'testing $ echo hi\n',
        'crud\n',
    ])

    expected = {
        'action': 'shell',
        'cwd': 'testing',
        'target': 'echo hi',
        'content': 'crud\n',
        'indent': 4,
        'line': None,
    }

    assert cmd == expected


@pytest.fixture
def example(tmpdir):
    p = tmpdir.join("example.txt")
    p.write(textwrap.dedent("""\
        a simple test function call with one argument factory
        ==============================================================

        the function argument.  Let's look at a simple self-contained
        example that you can put into a test module::

            # content of: test_simplefactory.py
            def pytest_funcarg__myfuncarg(request):
                return 42

            def test_function(myfuncarg):
                assert myfuncarg == 17

        .. code-block:: bash

            $ py.test test_simplefactory.py

            output should be here - but thats nice for testing

        the end.
    """))
    return p


def test_simple_new_content(tmpdir):

    needed_update, = check_file(
        name='example',
        content=simple.splitlines(True),
        tmpdir=str(tmpdir),
        normalize=[],
    )

    expected_update = {
        'action': 'shell',
        'cwd': None,
        'file': 'example',
        'target': 'echo hi',
        'content': 'oh no\n',
        'new_content': 'hi\n',
        'indent': 4,
        'line': 2,
    }
    print(needed_update)
    assert needed_update == expected_update


def test_single_update():

    update = {
        'action': 'shell',
        'target': 'echo hi',
        'content': 'oh no\n',
        'new_content': 'hi\n',
        'indent': 4,
        'line': 2,
    }

    corrected = correct_content(simple.splitlines(True), [update])
    assert corrected == simple_corrected.splitlines(True)


def test_actions_of(tmpdir, example):

    actions = list(parse_actions(example.readlines()))

    interesting = [a_c_t(x) for x in actions]
    expected = [
        ('write', 'test_simplefactory.py'),
        ('shell', 'py.test test_simplefactory.py')]

    assert interesting == expected


def test_check_file(tmpdir, example):

    check_file(
        name='test.txt',
        content=example.readlines(),
        tmpdir=str(tmpdir),
        normalize=[],
    )
    assert tmpdir.join('test_simplefactory.py').check()


def test_main_no_update(tmpdir, example, run):
    run(
        example,
        update=False,
        rootdir=str(tmpdir),
    )
    # check for the created tmpdir

    assert tmpdir.join('example.txt-1').check(dir=1)


def test_empty_update(tmpdir, run):
    simple_fp = tmpdir.join('simple.txt')
    simple_fp.write("")
    run(
        simple_fp,
        update=True,
        rootdir=str(tmpdir),
    )
    corrected = simple_fp.read()
    assert corrected == ""


def test_main_update(tmpdir, run):
    simple_fp = tmpdir.join('simple.txt')
    simple_fp.write(simple)
    run(
        simple_fp,
        update=True,
        rootdir=str(tmpdir),
    )
    corrected = simple_fp.read()

    assert corrected == simple_corrected


def test_docfile_chdir(tmpdir, monkeypatch):

    tmpdir.ensure('nested/file').write('some text\n')
    monkeypatch.chdir(tmpdir)
    filename = str(tmpdir.join('test.txt'))
    content = [
        'some shell test\n',
        '  nested $ cat file\n',
        '  some other text\n',
    ]

    action, = parse_actions(content, file=filename)
    excpected_action = {
        'action': 'shell',
        'content': 'some other text\n',
        'cwd': 'nested',
        'file': filename,
        'indent': 2,
        'line': 1,
        'target': 'cat file',
    }
    assert action == excpected_action

    needed_updates = check_file(
        name='example.txt',
        content=content,
        tmpdir=str(tmpdir.join('tmp')),
        normalize=[],)

    assert needed_updates
    # is it copied?
    assert tmpdir.join('tmp/nested/file').check()

def test_parsing_problem(tmpdir, run):
    simple_fp = tmpdir.join('index.txt')
    simple_fp.write(example_index)

    run(
        simple_fp,
        update=True,
        rootdir=str(tmpdir),
    )
    corrected = simple_fp.read()
    assert corrected == example_index

example_index = """
py.test: no-boilerplate testing with Python
==============================================

.. toctree::
"""


def test_wipe(tmpdir):
    fp = tmpdir.join('simple.txt')
    fp.write('.. regendoc:wipe\n')
    check_file(
        name='test',
        content=fp.readlines(),
        tmpdir=str(tmpdir),
        normalize=[],
    )

    assert not tmpdir.listdir()



def dedent(line, last_indent):
    if last_indent is not None:
        if line[:last_indent].isspace():
            return last_indent, line[last_indent:]
    stripped = line.lstrip(' ')
    return len(line) - len(stripped), stripped


def blocks(lines):
    result = []
    firstline = None
    last_indent = None
    items = []
    for lineno, line in enumerate(lines):

        indent, rest = dedent(line, last_indent)

        if last_indent is None:
            last_indent = indent

        if firstline is None:
            firstline = lineno

        if indent != last_indent:
            if items[0] == '\n':
                del items[0]
                firstline += 1
            if items and items[-1] == '\n':
                del items[-1]
            result.append((last_indent, firstline, items))
            items = [rest]
            last_indent = indent
            firstline = lineno

        else:
            last_indent = indent
            items.append(rest or '\n')

    else:
        try:
            result.append((indent, firstline, items))
        except UnboundLocalError:
            pass
    return result


def correct_content(content, updates):

    lines = list(content)
    for update in reversed(updates):
        line = update['line']
        old_lines = len(update['content'].splitlines())
        indent = ' ' * update['indent']
        new_lines = [indent + _line
                     for _line in update['new_content'].splitlines(1)]
        lines[line + 1:line + old_lines + 1] = new_lines

    return lines


def classify(lines, indent=4, line=None):
    if not lines:
        return {'action': None}
    first = lines[0]
    content = ''.join(lines[1:])

    def at(action, target, cwd=None):
        return {
            'action': action,
            'cwd': cwd,
            'target': target,
            'content': content,
            'indent': indent,
            'line': line,
        }

    if first.startswith('# content of'):
        target = first.strip().split()[-1]
        return at('write', target)
    elif first[0] == '$':
        cmd = first[1:].strip()
        return at('shell', cmd)
    elif ' $ ' in first:
        cwd, target = first.split(' $ ')
        target = target.strip()
        return at('shell', target, cwd)
    elif not indent and any(x.strip() == '.. regendoc:wipe' for x in lines):
        return {'action': 'wipe'}

    return at(None, first)


def parse_actions(lines, **kw):
    for indent, line, data in blocks(lines):
        mapping = classify(lines=data, indent=indent, line=line)
        mapping.update(kw)
        if mapping['action']:  # None if no idea
            yield mapping



from setuptools import setup

args = dict(
    name='regendoc',
    use_scm_version=True,
    description='a tool to check/update simple file/shell '
                'examples in documentation',
    url='http://bitbucket.org/pytest-dev/regendoc/',
    author='Ronny Pfannschmidt',
    author_email='opensource@ronnypfannschmidt.de',
    packages=['regendoc'],
    entry_points={
        'console_scripts': [
            'regendoc = regendoc:main',
        ]},
    install_requires=[
        'click',
    ],
    setup_requires=[
        'setuptools_scm',
    ]
)

if __name__ == '__main__':
    setup(**args)
